use crate::ast::named::Named;
use crate::ast::{CompilationUnit, Identifier};
use crate::diagnostic::DmDiagnostic;
use crate::name_analysis::fqn_context::FqnContext;
use crate::name_analysis::gather::gather_module_level_declaration;
use crate::name_analysis::resolve::resolve_module_level_declaration;
use crate::name_analysis::symbol_table::SymbolTable;
use codespan_reporting::diagnostic::{Diagnostic, Label};

mod fqn_context;
mod gather;
mod resolve;
pub mod symbol;
pub mod symbol_table;

pub(self) struct DiagnosticsContainer {
    file_id: usize,
    diagnostics: Vec<DmDiagnostic>,
}

impl DiagnosticsContainer {
    pub fn new(file_id: usize) -> DiagnosticsContainer {
        DiagnosticsContainer {
            file_id,
            diagnostics: vec![],
        }
    }

    pub fn add(&mut self, f: impl FnOnce(usize) -> DmDiagnostic) {
        self.diagnostics.push(f(self.file_id));
    }

    pub fn add_error_at_identifier(&mut self, msg: &str, identifier: &Identifier) {
        self.diagnostics.push(
            Diagnostic::error()
                .with_message(msg)
                .with_label(Label::primary(self.file_id, identifier.range)),
        );
    }
}

impl Into<Vec<DmDiagnostic>> for DiagnosticsContainer {
    fn into(self) -> Vec<DmDiagnostic> {
        self.diagnostics
    }
}

pub fn analyze_names(
    file_id: usize,
    compilation_unit: &mut CompilationUnit,
    symbol_table: &mut SymbolTable,
) -> Vec<DmDiagnostic> {
    let mut diagnostics = DiagnosticsContainer::new(file_id);

    let mut fqn_context = FqnContext::new();
    if let Some(namespace) = &compilation_unit.namespace {
        fqn_context.push(namespace.name().to_string());
    }

    for declaration in &mut compilation_unit.declarations {
        gather_module_level_declaration(
            declaration,
            symbol_table,
            &mut fqn_context,
            &mut diagnostics,
        );
    }

    assert_eq!(symbol_table.current_scope_id(), 0);

    for declaration in &mut compilation_unit.declarations {
        resolve_module_level_declaration(declaration, symbol_table, &mut diagnostics);
    }

    diagnostics.into()
}
