use crate::name_analysis::symbol::Symbol;
use std::collections::HashMap;
use std::fmt::Display;

pub struct Scope {
    parent: Option<usize>,
    symbols: HashMap<String, Symbol>,
    debug_name: String,
}

#[derive(Default)]
pub struct SymbolTable {
    scopes: Vec<Scope>,
    current_scope_id: usize,
}

/// Contains a vec of scopes, like a flattened tree
impl SymbolTable {
    pub fn new() -> Self {
        let mut t = SymbolTable::default();
        t.scopes.push(Scope {
            parent: None,
            symbols: HashMap::new(),
            debug_name: String::from("Global Scope"),
        });
        t.current_scope_id = 0;
        t
    }

    pub fn current_scope_id(&self) -> usize {
        self.current_scope_id
    }

    pub fn push_scope(&mut self, debug_name: &str) {
        let id = self.scopes.len();
        self.scopes.push(Scope {
            parent: Some(self.current_scope_id),
            symbols: HashMap::new(),
            debug_name: debug_name.to_string(),
        });
        self.current_scope_id = id;
    }

    pub fn pop_scope(&mut self) {
        if let Some(parent_id) = self.scopes[self.current_scope_id].parent {
            self.current_scope_id = parent_id;
        }
    }

    pub fn insert(&mut self, name: String, symbol: Symbol) -> Result<(), String> {
        if let Some(current_symbol) = self.scopes[self.current_scope_id].symbols.get(&name) {
            Err(format!(
                "Symbol '{}' already defined in current scope.",
                current_symbol.name()
            ))
        } else {
            self.scopes[self.current_scope_id]
                .symbols
                .insert(name, symbol);
            Ok(())
        }
    }

    pub fn lookup(&self, name: &str, scope_id: usize) -> Result<&Symbol, String> {
        let mut scope_opt = Some(&self.scopes[scope_id]);
        while let Some(scope) = scope_opt {
            if let Some(symbol) = scope.symbols.get(name) {
                return Ok(symbol);
            }
            scope_opt = if let Some(parent_id) = scope.parent {
                Some(&self.scopes[parent_id])
            } else {
                None
            };
        }
        Err(format!("Symbol '{}' not found in current scope.", name))
    }
}

impl Display for SymbolTable {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        for (i, scope) in self.scopes.iter().enumerate() {
            writeln!(f, "Scope {} {}", i, scope.debug_name)?;
            for (name, symbol) in &scope.symbols {
                writeln!(f, "  {}({})", name, symbol)?;
            }
        }
        Ok(())
    }
}
